"""
Common filtering logic
"""
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict, Any

from suricatalog.time import DEFAULT_TIMESTAMP_10M_AGO, parse_timestamp, to_utc


class BaseFilter(ABC):
    """
    Abstract filter
    """

    @abstractmethod
    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        Needs to be overridden with real logic
        :param data:
        :return:
        """
        raise NotImplementedError()


class AlwaysTrueFilter(BaseFilter):
    """
    This filter is always true
    """

    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        Always true implementation
        :param data:
        :return:
        """
        return True


class OnlyAlertsFilter(BaseFilter):
    """
    Filter only alerts
    """

    def __init__(self):
        """
        Constructor
        """
        self._timestamp = DEFAULT_TIMESTAMP_10M_AGO

    @property
    def timestamp(self):
        """
        Get timestamp
        :return:
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: datetime):
        """
        Set timestamp
        :param timestamp:
        :return:
        """
        if not timestamp:
            raise ValueError("Missing timestamp")
        if not self.timestamp.tzinfo:
            raise ValueError(f"{timestamp} has not TimeZone information")
        self._timestamp = timestamp

    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        Filter events based on time and only alerts
        :param data:
        :return:
        """
        try:
            event_timestamp = parse_timestamp(data['timestamp'])
            if event_timestamp <= self._timestamp:
                return False
            if 'event_type' in data and 'alert' == data['event_type']:
                return True
            return False
        except ValueError:
            return False


class NXDomainFilter(BaseFilter):
    """
    Filter for DNS code
    """

    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        tail -f eve.json|jq -c 'select(.dns.rcode=="NXDOMAIN")'
        """
        if 'dns' in data and 'rcode' in data['dns'] and data['dns']['rcode'] == 'NXDOMAIN':
            return True
        return False


class WithPrintablePayloadFilter(BaseFilter):
    """
    Show only records with a printable payload
    """

    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        cat ~/SuricataLog/test/eve.json | jq -r -c 'select(.event_type=="alert")|.payload'|base64 --decode
        :param data:
        :return:
        """
        if 'event_type' in data and 'alert' == data['event_type']:
            payload_printable = 'payload_printable' in data and data['payload_printable'] and data[
                'payload_printable'] != 'null'
            has_payload = 'payload' in data and data['payload'] and data['payload'] != 'null'
            if payload_printable or has_payload:
                return True
        return False


class TimestampFilter(BaseFilter):
    """
    Filter for events based on timestamp
    """

    def __init__(self):
        self._timestamp = DEFAULT_TIMESTAMP_10M_AGO

    @property
    def timestamp(self):
        """
        Get timestamps
        :return:
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: datetime):
        """"
        Set timestamp
        """
        if not timestamp:
            raise ValueError("Missing timestamp")
        if not timestamp.tzinfo:
            raise ValueError(f"{timestamp} has not TimeZone information")
        self._timestamp = timestamp

    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        Filter events on a given timestamp
        :param data:
        :return:
        """
        try:
            event_timestamp = parse_timestamp(data['timestamp'])
            if not event_timestamp.tzinfo:
                event_timestamp = to_utc(event_timestamp)
            if event_timestamp <= self._timestamp:
                return False
        except TypeError as te:
            if not self._timestamp.tzinfo:
                raise TypeError(f"today timestamp={self._timestamp} has not TimeZone information.") from te
        except ValueError:
            return False
        return True


class WithPayloadFilter(BaseFilter):
    """
    Filter records with any payload
    """

    def accept(self, data: Dict[Any, Any]) -> bool:
        """
        cat ~/SuricataLog/test/eve.json | jq -r -c 'select(.event_type=="alert")|.payload'|base64 --decode
        :param data:
        :return:
        """
        if 'event_type' in data and 'alert' == data['event_type']:
            payload = 'payload' in data and data['payload'] and data[
                'payload'] != 'null'
            if payload:
                return True
        return False



"""
Common logic to handle copying data to the clipboard.
Depending on how much data you are displaying, these may can crash the application.
"""
import json
from typing import Tuple, Any, List

import pyperclip
from textual.widgets import RichLog, DataTable, Digits


def copy_from_richlog(rich_log: RichLog) -> Tuple[str, int]:
    """
    Extract data from rich log.
    Messages are not properly formatted as JSON, they are copied 'as-is'
    :param rich_log:
    :return:
    """
    if rich_log is None:
        return "", 0
    text_to_copy = '\n'.join([strip.text.strip() for strip in rich_log.lines])
    pyperclip.copy(text_to_copy)
    return text_to_copy, len(text_to_copy)


def copy_from_digits(digits: Digits) -> Tuple[str, int]:
    """
    Extract data from digits
    :param digits:
    :return:
    """
    if digits is None:
        return "", 0
    text_to_copy = digits.value
    pyperclip.copy(text_to_copy)
    return text_to_copy, len(text_to_copy)


def copy_from_table(data_table: DataTable) -> Tuple[List[Any], int] | Tuple[str, int]:
    """
    Extract data from table
    :param data_table:
    :return:
    """
    if data_table is None:
        return [], 0
    rows = []
    columns = []
    for key, column in data_table.columns.items():
        columns.append(column.label.plain)
    rows.append(columns)
    for key in data_table.rows:
        row = data_table.get_row(key)
        rows.append(row)
    text_to_copy = json.dumps(rows, indent=2, sort_keys=True)
    pyperclip.copy(text_to_copy)
    return text_to_copy, len(text_to_copy)



"""
Unit tests for payloads
"""

import json
import unittest
from unittest import TestCase
from test.test_log import BASEDIR
from suricatalog.filter import WithPrintablePayloadFilter, WithPayloadFilter, NXDomainFilter


class FilterTestCase(TestCase):
    """
    Unit test for filters
    """

    flow_ids = [
        1117051772115445,
        1117051772115445,
        879572199993573,
        1209610707920645,
        1101326487668301,
        248862673879386,
        248862673879386,
        1707519430412687,
        1707519430412687,
        879864496420835,
        1083777546714183,
        1357514933978659,
        879864496420835,
        206845708466045,
        653057280496748,
        232044371766299,
        833894710498356,
        833894710498356,
        667366924188955,
        94511204925000,
        41564274445680,
        277752553385435,
        806136472100536,
        611798819162769,
        749924187348801,
        1248865274295334,
        784224188217537,
        2132655805578936,
        2250553857023101,
        1980577284838437,
        1208016640841156,
        1980577284838437,
        1255871085730631,
        1296817044046536,
        1262262924962648,
        1211568756515917,
        636551925240335,
        657941680360877,
        644785005397810,
        249279597341217,
        201901383894597,
        242479979619734,
        394539153315441,
        244035746600868,
        278410102952556,
        685462364109190,
        1116734859441197,
        1116734859441197,
        1929642955916282,
        784369590587384,
        1911867894557238,
        1387833132456197,
        1616909837011682,
        1690259766491386,
        1616909837011682,
        1464936998127434,
        1579061810494998,
        1464936998127434,
        1579061810494998,
        2024483847095705,
        1692920326018047,
        2240848756896914,
        1684109579010675,
        1605766509821287,
        1736380214178990,
        1684109579010675,
        2198502421252463,
        2198502421252463,
        808114184483419,
        719606258986631,
        808114184483419,
        719606258986631,
        947551783206878,
        947551783206878,
        1431171568122864,
        382125840958721,
        382125840958721,
        2133828637592443,
        2109669390050495,
        2135326759364959,
        1913267474427766,
        1838515100344931,
        1915756617234611,
        1916415357571604,
        487984813931697,
        296466617085665,
        408302945206078,
        348636333008022,
        284724912486502,
        284724912486502,
        1655729925536716,
        733464559355118,
        348832191139035,
        1156650674014842,
        1640122311458474,
        356383397518687,
        1156650674014842,
        792788483047503,
        369725903275355,
        1105447458998123,
        1105447458998123,
        943516941319943,
        1303316325186687,
        644785005397810,
        97676645382088,
        296466617085665,
        2109669390050495,
        368053091258192,
        249279597341217,
        1395209664101095,
        667366924188955,
        2133828637592443,
        408302945206078,
        784369590587384,
        1690259766491386,
        348636333008022,
        242479979619734,
        1605766509821287
    ]

    def test_with_printable_payload_filter(self):
        """
        Payload test
        :return:
        """
        data_filter = WithPrintablePayloadFilter()
        no_payload = json.loads('{"timestamp":"2022-02-08T16:32:14.900292+0000","flow_id":879564658970874,'
                                '"pcap_cnt":9146,"event_type":"alert","src_ip":"52.96.222.130","src_port":25,'
                                '"dest_ip":"10.2.8.102","dest_port":49880,"proto":"TCP","metadata":{"flowints":{'
                                '"applayer.anomaly.count":1}},"alert":{"action":"allowed","gid":1,'
                                '"signature_id":2260002,"rev":1,"signature":"SURICATA Applayer Detect protocol only one '
                                'direction","category":"Generic Protocol Command Decode","severity":3},"smtp":{"helo":"['
                                '173.166.146.112]"},"app_proto":"smtp","app_proto_tc":"failed",'
                                '"flow":{"pkts_toserver":3,"pkts_toclient":3,"bytes_toserver":198,"bytes_toclient":276,'
                                '"start":"2022-02-08T16:32:14.659706+0000"},"payload":"","payload_printable":"",'
                                '"stream":0,"packet":"AAgCHEevIOUqtpPxCABFAAAohfUAAIAGj5A0YN6CCgIIZgAZwth27Ik6+soFbVAQ'
                                '+vDMSAAA","packet_info":{"linktype":1},"host":"raspberrypi"}')
        self.assertFalse(data_filter.accept(no_payload))
        payload = json.loads('{"timestamp":"2022-02-08T16:32:20.491791+0000","flow_id":879564658970874,'
                             '"pcap_cnt":9254,"event_type":"alert","src_ip":"52.96.222.130","src_port":25,'
                             '"dest_ip":"10.2.8.102","dest_port":49880,"proto":"TCP","metadata":{"flowints":{'
                             '"applayer.anomaly.count":1,"smtp.anomaly.count":1}},"tx_id":0,'
                             '"alert":{"action":"allowed","gid":1,"signature_id":2220000,"rev":1,'
                             '"signature":"SURICATA SMTP invalid reply","category":"Generic Protocol Command Decode",'
                             '"severity":3},"smtp":{"helo":"[173.166.146.112]"},"app_proto":"smtp",'
                             '"app_proto_tc":"failed","flow":{"pkts_toserver":9,"pkts_toclient":14,'
                             '"bytes_toserver":570,"bytes_toclient":1148,"start":"2022-02-08T16:32:14.659706+0000"},'
                             '"payload":"TV0NCg==","payload_printable":"220 PH0P220CA0030.outlook.office365.com",'
                             '"stream":1,"packet":"AAgCH","packet_info":{"linktype":1},"host":"raspberrypi"}')
        self.assertTrue(data_filter.accept(payload))

    def test_with_printable_payload_filter_from_file(self):
        """
        Test printable payload
        :return:
        """
        data_filter = WithPrintablePayloadFilter()
        # 84 of these payloads have a 'null',so they are expected to be filtered
        accepted_cnt = 34
        counted = 0
        with open(BASEDIR.joinpath("eve-2.json"), 'rt', encoding='utf-8') as eve_file:
            for event in eve_file:
                payload = json.loads(event)
                accepted = data_filter.accept(payload)
                """
                Validation is tricky, as the payload is split across several alerts. For example:
                {"flow_id":1117051772115445,"payload":null,"alert":{"action":"allowed","gid":1,"signature_id":2260002,"rev":1,"signature":"SURICATA,
                ...
                {"flow_id":1117051772115445,"payload":
                "MjIwIHNtdHAubWFpbC55YWhvby5jb20gRVNNVFAgcmVhZHkNCjI1MC1rdWJlbm9kZTUxNS5tYWlsLXByb2QxLm9tZWdhLmdxMS55YWhvby\
                5jb20gSGVsbG8gWzE3My4xNjYuMTQ2LjExMl0gWzE3My4xNjYuMTQ2LjExMl0pDQoyNTAtRU5IQU5DRURTVEFUVVNDT0RFUw0KMjUwLThCSVRNSU1FDQoy\
                NTAtU0laRSA0MTY5NzI4MA0KMjUwIEFVVEggUExBSU4gTE9HSU4gWE9BVVRIMiBPQVVUSEJFQVJFUg0KMzM0IFZYTmxjbTVoYldVNg0KMzM0IFVHRnpjM2R\
                2Y21RNg0KNTM1IDUuNy4wICgjQVVUSDAwNSkgVG9vIG1hbnkgYmFkIGF1dGggYXR0ZW1wdHMuDQo=",\
                "alert":{"action":"allowed","gid":1,"signature_id":2220000,"rev":1,"signature":"SURICATA,
                So we must manually ensure there is indeed a payload
                """
                if 'flow_id' in payload:
                    flow_id = payload['flow_id']
                    if flow_id in self.flow_ids:
                        if 'payload' in payload and not payload['payload'] == 'null':
                            self.assertTrue(accepted, payload)
                            counted += 1
                    else:
                        self.assertFalse(accepted, payload)
        self.assertEqual(accepted_cnt, counted)

    def test_with_payload_filter_from_file(self):
        """
        Test printable payload
        :return:
        """
        data_filter = WithPayloadFilter()
        # 84 of these payloads have a 'null',so they are expected to be filtered
        accepted_cnt = 34
        counted = 0
        with open(BASEDIR.joinpath("eve-2.json"), 'rt', encoding='utf-8') as eve_file:
            for event in eve_file:
                payload = json.loads(event)
                accepted = data_filter.accept(payload)
                """
                Validation is tricky, as the payload is split across several alerts. For example:
                {"flow_id":1117051772115445,"payload":null,"alert":{"action":"allowed","gid":1,"signature_id":2260002,"rev":1,"signature":"SURICATA,
                ...
                {"flow_id":1117051772115445,"payload":
                "MjIwIHNtdHAubWFpbC55YWhvby5jb20gRVNNVFAgcmVhZHkNCjI1MC1rdWJlbm9kZTUxNS5tYWlsLXByb2QxLm9tZWdhLmdxMS55YWhvby\
                5jb20gSGVsbG8gWzE3My4xNjYuMTQ2LjExMl0gWzE3My4xNjYuMTQ2LjExMl0pDQoyNTAtRU5IQU5DRURTVEFUVVNDT0RFUw0KMjUwLThCSVRNSU1FDQoy\
                NTAtU0laRSA0MTY5NzI4MA0KMjUwIEFVVEggUExBSU4gTE9HSU4gWE9BVVRIMiBPQVVUSEJFQVJFUg0KMzM0IFZYTmxjbTVoYldVNg0KMzM0IFVHRnpjM2R\
                2Y21RNg0KNTM1IDUuNy4wICgjQVVUSDAwNSkgVG9vIG1hbnkgYmFkIGF1dGggYXR0ZW1wdHMuDQo=",\
                "alert":{"action":"allowed","gid":1,"signature_id":2220000,"rev":1,"signature":"SURICATA,
                So we must manually ensure there is indeed a payload
                """
                if 'flow_id' in payload:
                    flow_id = payload['flow_id']
                    if flow_id in self.flow_ids:
                        if 'payload' in payload and not payload['payload'] == 'null':
                            self.assertTrue(accepted, payload)
                            counted += 1
                    else:
                        self.assertFalse(accepted, payload)
        self.assertEqual(accepted_cnt, counted)

    def test_nxdomain(self):
        """
        Test NX DNS app
        :return:
        """
        data_filter = NXDomainFilter()
        payload = json.loads('{"timestamp":"2022-02-23T19:04:05.606882+0000","flow_id":1452066252079368,'
                             '"pcap_cnt":85538,"event_type":"dns","src_ip":"172.16.0.170","src_port":58529,'
                             '"dest_ip":"172.16.0.52","dest_port":53,"proto":"UDP","dns":{"version":2,"type":"answer",'
                             '"id":46693,"flags":"8583","qr":true,"aa":true,"rd":true,"ra":true,'
                             '"rrname":"wpad.sunnystation.com","rrtype":"A","rcode":"NXDOMAIN","authorities":[{'
                             '"rrname":"sunnystation.com","rrtype":"SOA","ttl":3600,'
                             '"soa":{"mname":"sunnystation-dc.sunnystation.com","rname":"hostmaster.sunnystation.com",'
                             '"serial":32,"refresh":900,"retry":600,"expire":86400,"minimum":3600}}]},'
                             '"host":"raspberrypi"}')
        self.assertTrue(data_filter.accept(payload))
        payload = json.loads('{"timestamp":"2022-02-08T15:08:22.266264+0000","flow_id":127219202876190,'
                             '"pcap_cnt":4196,"event_type":"tls","src_ip":"10.2.8.102","src_port":49798,'
                             '"dest_ip":"185.248.140.40","dest_port":443,"proto":"TCP","tls":{"subject":"C=GB, '
                             'ST=London, L=London, O=Global Security, OU=IT Department, CN=example.com",'
                             '"issuerdn":"C=GB, ST=London, L=London, O=Global Security, OU=IT Department, '
                             'CN=example.com","serial":"0E:B4:42:BB:46:DA:06:75:CF:4B:CA:7B:2B:E8:C2:60:0B:18:32:17",'
                             '"fingerprint":"28:c1:5a:60:d5:cb:a4:89:4b:07:c6:4c:d8:67:e7:f1:76:ef:da:aa",'
                             '"version":"TLS 1.2","notbefore":"2022-01-31T20:50:00","notafter":"2023-01-31T20:50:00",'
                             '"ja3":{"hash":"51c64c77e60f3980eea90869b68c58a8","string":"771,'
                             '49196-49195-49200-49199-49188-49187-49192-49191-49162-49161-49172-49171-157-156-61-60'
                             '-53-47-10,10-11-13-35-23-65281,29-23-24,0"},"ja3s":{'
                             '"hash":"ec74a5c51106f0419184d0dd08fb05bc","string":"771,49200,65281-11-35-23"}},'
                             '"host":"raspberrypi"}')
        self.assertFalse(data_filter.accept(payload))


if __name__ == '__main__':
    unittest.main()



"""
Unit test for logging
"""
import json
import unittest
from datetime import datetime
from pathlib import Path

import pytz

from suricatalog.filter import TimestampFilter, AlwaysTrueFilter
from suricatalog.log import get_events_from_eve
from suricatalog.time import to_utc, parse_timestamp

BASEDIR = Path(__file__).parent


class SuricataLogTestCase(unittest.TestCase):
    """
    Unit test for suricata log common tests
    """
    eve_list = []
    old_date = datetime(
        year=2021,
        day=2,
        month=8,
        tzinfo=pytz.UTC
    )

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup data loading
        :return:
        """
        with open(BASEDIR.joinpath("eve.json"), 'rt', encoding='utf-8') as eve_file:
            for event in eve_file:
                SuricataLogTestCase.eve_list.append(json.loads(event))

    def test_to_utc(self):
        """
        Test UCT datetime handling
        :return:
        """
        naive = datetime.now()
        non_naive = datetime(
            year=2024,
            day=2,
            month=2,
            tzinfo=pytz.UTC
        )
        dates = [
            naive,
            non_naive
        ]
        for test_date in dates:
            ts = to_utc(test_date)
            self.assertTrue(ts)
            self.assertIsInstance(ts, datetime)
            self.assertIsNotNone(ts.tzinfo)

    def test_parse_timestamp(self):
        """
        Test timestamp parsing
        :return:
        """
        invalid = 'XXX-02-08T16:32:14.900292'
        naive = datetime.now()
        non_naive = datetime(
            year=2024,
            day=2,
            month=2,
            tzinfo=pytz.UTC
        )
        dates = [
            '2022-02-08T16:32:14.900292+0000',
            '2022-02-08 16:32:14.900292+0000',
            '2022-02-08T16:32:14.900292',
            naive,
            non_naive,
            invalid
        ]
        for idx, test_date in enumerate(dates):
            try:
                ts = parse_timestamp(test_date)
                self.assertTrue(ts)
                self.assertIsInstance(ts, datetime)
                self.assertIsNotNone(ts.tzinfo)
            except ValueError:
                if idx == 6:
                    self.fail(f"Was supposed to fail with an invalid timestamp: {invalid}")
                else:
                    pass

    def test_timestamp_filter(self):
        """
        Test timestamp filtering
        :return:
        """
        timestamp_filter = TimestampFilter()
        timestamp_filter.timestamp = SuricataLogTestCase.old_date
        self.assertTrue(
            timestamp_filter.accept(data=SuricataLogTestCase.eve_list[0])
        )
        self.assertTrue(
            timestamp_filter.accept(data=SuricataLogTestCase.eve_list[132])
        )

    def test_get_alerts(self):
        """
        Test get alerts
        :return:
        """
        timestamp_filter = TimestampFilter()
        timestamp_filter.timestamp = SuricataLogTestCase.old_date
        all_alerts = [x for x in get_events_from_eve(
            eve_files=[BASEDIR.joinpath("eve.json")],
            data_filter=timestamp_filter
        ) if x['event_type'] == 'alert']
        self.assertIsNotNone(all_alerts)
        self.assertEqual(275, len(all_alerts))
        self.assertEqual('SURICATA Applayer Detect protocol only one direction', all_alerts[90]['alert']['signature'])

    def test_get_all_events(self):
        """
        Test get all event, unfiltered
        :return:
        """
        always_true_filter = AlwaysTrueFilter()
        all_events = list(get_events_from_eve(
            eve_files=[BASEDIR.joinpath("eve.json")],
            data_filter=always_true_filter
        ))
        self.assertIsNotNone(all_events)
        self.assertListEqual(SuricataLogTestCase.eve_list, all_events)


if __name__ == '__main__':
    unittest.main()



"""
Unit test for reports
"""
import json
import unittest
from unittest import IsolatedAsyncioTestCase
from test.test_log import BASEDIR
from suricatalog.report import AggregatedFlowProtoReport, HostDataUseReport, TopUserAgents


if __name__ == '__main__':
    unittest.main()


class ReportTestCase(IsolatedAsyncioTestCase):
    """
    Report test case
    """
    eve_list = []
    netflow_eve_list = []

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup data for testing
        :return:
        """
        with open(BASEDIR.joinpath("eve.json"), 'rt', encoding='utf-8') as eve_file:
            for event in eve_file:
                ReportTestCase.eve_list.append(json.loads(event))
        with open(BASEDIR.joinpath("eve_udp_flow.json"), 'rt', encoding='utf-8') as eve_file:
            for event in eve_file:
                ReportTestCase.netflow_eve_list.append(json.loads(event))

    async def test_aggregated_flow_proto_report(self):
        """
        Test flow report
        :return:
        """
        can_rep = AggregatedFlowProtoReport()
        for alert in ReportTestCase.eve_list:
            await can_rep.ingest_data(alert)
        self.assertIsNotNone(can_rep.port_proto_count)
        self.assertEqual(119, len(can_rep.port_proto_count.keys()))
        self.assertEqual(336, can_rep.port_proto_count[('TCP', 443)])

    async def test_host_data_use_report(self):
        """
        Test data use report
        :return:
        """
        hr = HostDataUseReport()
        for netflow_event in ReportTestCase.netflow_eve_list:
            await hr.ingest_data(netflow_event, '224.0.0.251')
        self.assertEqual(153339, hr.bytes)
        hr2 = HostDataUseReport()
        for netflow_event in ReportTestCase.netflow_eve_list:
            await hr2.ingest_data(netflow_event, '127.0.0.1')
            await hr.ingest_data(netflow_event, '127.0.0.1')
        self.assertEqual(0, hr2.bytes)
        self.assertEqual(153339, hr.bytes)

    async def test_top_user_agents(self):
        """
        Test top user agents
        :return:
        """
        tua = TopUserAgents()
        for event in ReportTestCase.eve_list:
            await tua.ingest_data(event)

        self.assertIsNotNone(tua.agents)
        self.assertTrue('test' in tua.agents)
        self.assertEqual(10, len(tua.agents))



"""
Collection of canned reports

"""
import dataclasses
from typing import Dict, Any, Tuple


@dataclasses.dataclass
class AggregatedFlowProtoReport:
    """
    FLow protocol report details
    """
    port_proto_count: Dict[Tuple[str, int], int] = dataclasses.field(default_factory=dict)

    async def ingest_data(self, data: Dict[Any, Any]) -> None:
        """
        ports=$(cat $PWD/test/eve.json|jq -c 'select(.event_type=="flow")|[.proto, .dest_port]'|sort |uniq -c)
        echo $ports
        1043 ["UDP",53]
        336 ["TCP",443]
        313 ["TCP",587]
        216 ["TCP",465]
        206 ["TCP",25]
        122 ["UDP",137]
        104 ["TCP",8080]
        70 ["TCP",88]
        61 ["TCP",80]
        59 ["TCP",389]

        To get only the top 10: echo $ports| sort -nr| head -n10
        :param data:
        :return:
        """
        if 'event_type' in data and 'flow' == data['event_type'] and 'dest_port' in data and 'proto' in data:
            port = data['dest_port'] if data['dest_port'] else ""
            proto_and_dest_port = (data['proto'], port)
            if proto_and_dest_port not in self.port_proto_count:
                self.port_proto_count[proto_and_dest_port] = 0
            self.port_proto_count[proto_and_dest_port] += 1


@dataclasses.dataclass
class HostDataUseReport:
    """
    Host data usage report
    """
    bytes: int = 0

    async def ingest_data(self, data: Dict[Any, Any], dest: str) -> None:
        """
        tail -n500000 /var/log/suricata/eve.json | \
        jq -s 'map(select(.event_type=="netflow" and .dest_ip=="224.0.0.251").netflow.bytes)|add'| /bin/numfmt --to=iec
        1.6M
        :param data:
        :param dest:
        :return:
        """
        if 'event_type' in data and 'netflow' == data['event_type'] and 'dest_ip' in data and dest == data['dest_ip']:
            self.bytes += data['netflow']['bytes']


class TopUserAgents:
    """
    Replicate tutorial top user agents query with jq.
    """

    agents: Dict[str, int] = {}

    async def ingest_data(self, data: Dict[Any, Any]) -> None:
        """
        cat eve.json | jq -s '[.[]|.http.http_user_agent]|group_by(.)|map({key:.[0],value:(.|length)})|from_entries'

        This particular rule is quite brittle, if data is incomplete you will get the following error:

        ```json
        jq: error (at <stdin>:14914): Cannot use null (null) as object key
        ```

        Instead of something like this:
        ```json
        {
            "Microsoft NCSI": 5,
            "Microsoft-CryptoAPI/6.1": 2,
            "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 10.0; Win64; x64; Trident/7.0; .NET4.0C; .NET4.0E; .NET CLR 2.0.50727;
            .NET CLR 3.0.30729; .NET CLR 3.5.30729)": 2,
            "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.1; Win64; x64; Trident/7.0; .NET CLR 2.0.50727; SLCC2;
            .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; .NET4.0C; .NET4.0E)": 6,
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Trident/7.0; rv:11.0) like Gecko": 3,
            "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko": 2,
            "WinHTTP loader/1.0": 4,
            "WinHTTP sender/1.0": 2,
            "pwtyyEKzNtGatwnJjmCcBLbOveCVpc": 2,
            "test": 4
        }
        ```
        :param data:
        :return:
        """
        if 'event_type' in data and 'http' in data and 'http_user_agent' in data['http']:
            agent = data['http']['http_user_agent']
            if agent:
                if agent not in self.agents:
                    self.agents[agent] = 0
                self.agents[agent] += 1



"""
Alert applications
"""
import textwrap
import traceback
from pathlib import Path
from typing import Type, List, Any, Dict, Union

import pyperclip
from textual import on, work
from textual.app import App, ComposeResult, CSSPathType
from textual.driver import Driver
from textual.widgets import Footer, Header, DataTable

from suricatalog.clipboard import copy_from_table
from suricatalog.log import get_events_from_eve
from suricatalog.filter import BaseFilter
from suricatalog.providers import TableAlertProvider, TableColumns
from suricatalog.screens import DetailScreen, ErrorScreen


class BaseAlertApp(App):
    """
    Base application for alert applications, shared logic
    """

    def __init__(
            self,
            driver_class: Type[Driver] | None = None,
            css_path: CSSPathType | None = None,
            watch_css: bool = False,
    ):
        """
        Constructor
        :param driver_class:
        :param css_path:
        :param watch_css:
        """
        super().__init__(driver_class, css_path, watch_css)
        self.eve_files = None
        self.filter = None

    @staticmethod
    def __get_key_from_map__(map1: Dict[str, Any], keys: List[str]):
        """
        Return the first matching key from a map
        :param map1:
        :param keys:
        :return: Nothing if none of the keys are in the map
        """
        val = ""
        for key in keys:
            if key in map1:
                val = map1[key]
                break
        return val

    @staticmethod
    async def extract_from_alert(alert: Dict[str, Any]) -> Dict[str, Any]:
        """
        Extract alerts from event
        :param alert:
        :return:
        """
        timestamp = alert['timestamp']
        dest_port = str(BaseAlertApp.__get_key_from_map__(alert, ['dest_port']))
        dest_ip = BaseAlertApp.__get_key_from_map__(alert, ['dest_ip'])
        src_ip = BaseAlertApp.__get_key_from_map__(alert, ['src_ip'])
        src_port = str(BaseAlertApp.__get_key_from_map__(alert, ['src_port']))
        protocol = BaseAlertApp.__get_key_from_map__(alert, ['app_proto', 'proto'])
        severity = alert['alert']['severity']
        if 'signature' in alert:
            signature = alert.get('signature', '')
        else:
            signature = alert['alert'].get('signature', '')
        payload_printable = alert.get('payload_printable', '')
        return {
            "timestamp": timestamp,
            "dest_port": dest_port,
            "src_ip": src_ip,
            "dest_ip": dest_ip,
            "src_port": src_port,
            "protocol": protocol,
            "severity": severity,
            "signature": signature,
            "payload_printable": payload_printable
        }

    def set_filter(self, the_filter: BaseFilter):
        """
        Set filter for application
        :param the_filter:
        :return:
        """
        if not the_filter:
            raise ValueError("Filter is required")
        self.filter = the_filter

    def set_eve_files(self, eve_files: List[Path]):
        """
        Set eve files for application
        :param eve_files:
        :return:
        """
        if not eve_files:
            raise ValueError("One or more eve files is required")
        self.eve_files = eve_files


class TableAlertApp(BaseAlertApp):
    """
    Concrete implementation for alert application
    """
    BINDINGS = [
        ("q", "quit_app", "Quit"),
        ("c", "copy_table", "Copy to clipboard")
    ]
    ENABLE_COMMAND_PALETTE = True
    COMMANDS = App.COMMANDS | {TableAlertProvider}

    def __init__(
            self,
            driver_class: Type[Driver] | None = None,
            css_path: CSSPathType | None = None,
            watch_css: bool = False,
    ):
        """
        Constructor
        :param driver_class:
        :param css_path:
        :param watch_css:
        """
        super().__init__(driver_class, css_path, watch_css)
        self.events: Union[Dict[Dict[str, any]], Dict] = {}

    async def show_error(
            self,
            trace: Union[traceback.StackSummary, None],
            reason: Union[str, None]
    ) -> None:
        """
        Show error on special screen
        :param trace:
        :param reason:
        :return:
        """
        error_src = ErrorScreen(trace=trace, reason=reason)
        await self.push_screen(error_src)

    def compose(self) -> ComposeResult:
        """
        Place TUI components
        :return:
        """
        yield Header()
        alerts_tbl = DataTable()
        alerts_tbl.show_header = True
        alerts_tbl.add_column(TableColumns.TIMESTAMP.name)
        alerts_tbl.add_column(TableColumns.SEVERITY.name)
        alerts_tbl.add_column(TableColumns.SIGNATURE.name)
        alerts_tbl.add_column(TableColumns.PROTOCOL.name)
        alerts_tbl.add_column(TableColumns.DESTINATION.name)
        alerts_tbl.add_column(TableColumns.SOURCE.name)
        alerts_tbl.add_column(TableColumns.PAYLOAD.name)
        alerts_tbl.zebra_stripes = True
        alerts_tbl.loading = True
        alerts_tbl.cursor_type = 'row'
        alerts_tbl.tooltip = textwrap.dedent("""
        Suricata alert details. Select a row to get full details.
        """)
        yield alerts_tbl
        yield Footer()

    @work(exclusive=False)
    async def on_mount(self) -> None:
        """
        Initialize TUI components
        :return:
        """
        alerts_tbl = self.query_one(DataTable)
        alert_cnt = 0
        try:
            events = get_events_from_eve(
                    data_filter=self.filter,
                    eve_files=self.eve_files
            )
            for event in events:
                if not self.filter.accept(event):
                    continue
                brief_data = await BaseAlertApp.extract_from_alert(event)
                timestamp = brief_data['timestamp']
                alerts_tbl.add_row(
                    timestamp,
                    brief_data['severity'],
                    brief_data['signature'],
                    brief_data['protocol'],
                    f"{brief_data['dest_ip']}:{brief_data['dest_port']}",
                    f"{brief_data['src_ip']}:{brief_data['src_port']}",
                    brief_data['payload_printable']
                )
                alert_cnt += 1
                self.events[timestamp] = event
            alerts_tbl.sub_title = f"Total alerts: {alert_cnt}"
            self.notify(
                title="Finish loading events",
                timeout=5,
                severity="information",
                message=textwrap.dedent(f"""
                Loaded {alert_cnt} messages.
                Click on a row to get more details, CTR+\\ to search
                """)
            )
            if alert_cnt:
                alerts_tbl.loading = False
            else:
                await self.show_error(reason=None, trace=None)
        except ValueError as ve:
            if hasattr(ve, 'reason'):
                reason = f"{ve}"
            elif hasattr(ve, 'message'):
                reason = ve.message
            else:
                reason = f"{ve}"
            self.log.info(f"Fatal error: {reason}")
            tb = traceback.extract_stack()
            self.log.info(tb)
            await self.show_error(trace=tb, reason=str(ve))

    @on(DataTable.HeaderSelected)
    def on_header_clicked(self, event: DataTable.HeaderSelected):
        """
        Handle click events on table
        :param event:
        :return:
        """
        alerts_tbl: DataTable = event.data_table
        alerts_tbl.sort(event.column_key)

    @on(DataTable.RowSelected)
    def on_row_clicked(self, event: DataTable.RowSelected) -> None:
        """
        Handle click on rows
        :param event:
        :return:
        """
        table = event.data_table
        row_key = event.row_key
        row = table.get_row(row_key)
        data = self.events[row[0]]
        runner_detail = DetailScreen(data=data)
        self.push_screen(runner_detail)

    def action_quit_app(self) -> None:
        """
        Handle exit action
        :return:
        """
        self.exit("Exiting Alerts now...")

    def action_copy_table(self) -> None:
        """
        Copy contents to the clipboard
        :return:
        """
        table_data = self.query_one(DataTable)
        try:
            _, ln = copy_from_table(table_data)
            self.notify(f"Copied {ln} characters!", title="Copied selection")
        except pyperclip.PyperclipException as exc:
            # Show a toast popup if we fail to copy.
            self.notify(
                str(exc),
                title="Clipboard error",
                severity="error",
            )



"""
Top user application related code
"""
from pathlib import Path
from typing import Type, List

from textual import work
from textual.app import App, ComposeResult, CSSPathType
from textual.driver import Driver
from textual.widgets import Header, RichLog, Footer
import pyperclip

from suricatalog import BASEDIR
from suricatalog.clipboard import copy_from_richlog
from suricatalog.filter import BaseFilter
from suricatalog.log import get_events_from_eve
from suricatalog.report import TopUserAgents


class TopUserApp(App):
    """
    Show top users
    """
    BINDINGS = [
        ("q", "quit_app", "Quit"),
        ("y,c", "copy_log", "Copy to clipboard")
    ]
    CSS_PATH = BASEDIR.joinpath('css').joinpath('canned.tcss')
    ENABLE_COMMAND_PALETTE = False

    def __init__(
            self,
            driver_class: Type[Driver] | None = None,
            css_path: CSSPathType | None = None,
            watch_css: bool = False,
            data_filter: BaseFilter = None,
            eve: List[Path] = None
    ):
        """
        Constructor
        :param driver_class:
        :param css_path:
        :param watch_css:
        :param data_filter:
        :param eve:
        """
        super().__init__(driver_class, css_path, watch_css)
        self.data_filter = data_filter
        self.eve_files = eve

    def action_quit_app(self) -> None:
        """
        Exit app
        :return:
        """
        self.exit("Exiting Top user now...")

    def action_copy_log(self) -> None:
        """
        Copy contents to the clipboard
        :return:
        """
        rich_log = self.query_one(RichLog)
        try:
            _, ln = copy_from_richlog(rich_log)
            self.notify(f"Copied {ln} characters!", title="Copied selection")
        except pyperclip.PyperclipException as exc:
            # Show a toast popup if we fail to copy.
            self.notify(
                str(exc),
                title="Clipboard error",
                severity="error",
            )

    def compose(self) -> ComposeResult:
        """
        Component placement
        :return:
        """
        yield Header()
        pretty = RichLog(
            id="agent",
            highlight=True,
            auto_scroll=True
        )
        pretty.loading = True
        yield pretty
        yield Footer()

    @work(exclusive=False)
    async def on_mount(self) -> None:
        """
        Populate TUI components with data
        :return:
        """
        top_user_agents = TopUserAgents()
        log = self.query_one("#agent", RichLog)
        log.loading = False
        for event in get_events_from_eve(
                eve_files=self.eve_files,
                data_filter=self.data_filter):
            await top_user_agents.ingest_data(event)
        log.write(top_user_agents.agents)



"""
One shot application related code
"""
import traceback
from pathlib import Path
from typing import Type, List

import pyperclip
from textual import work
from textual.app import App, ComposeResult, CSSPathType
from textual.driver import Driver
from textual.widgets import Header, RichLog, Footer

from suricatalog.clipboard import copy_from_richlog
from suricatalog.filter import BaseFilter
from suricatalog.log import get_events_from_eve
from suricatalog.screens import ErrorScreen


class OneShotApp(App):
    """
    One shot application, displays data based on filter
    """
    BINDINGS = [
        ("q", "quit_app", "Quit"),
        ("y,c", "copy_log", "Copy to clipboard")
    ]
    ENABLE_COMMAND_PALETTE = False

    def __init__(
            self,
            driver_class: Type[Driver] | None = None,
            css_path: CSSPathType | None = None,
            watch_css: bool = False,
            eve: List[Path] = None,
            data_filter: BaseFilter = None
    ):
        """
        Constructor
        :param driver_class:
        :param css_path:
        :param watch_css:
        :param eve:
        :param data_filter:
        """
        super().__init__(driver_class, css_path, watch_css)
        self.data_filter = data_filter
        self.eve = eve
        self.loaded = 0

    def action_quit_app(self) -> None:
        """
        Exit application
        :return:
        """
        self.exit(f"Exiting {self.title} now...")

    def action_copy_log(self) -> None:
        """
        Copy contents to the clipboard
        :return:
        """
        rich_log = self.query_one(RichLog)
        try:
            _, ln = copy_from_richlog(rich_log)
            self.notify(f"Copied {ln} characters!", title="Copied selection")
        except pyperclip.PyperclipException as exc:
            # Show a toast popup if we fail to copy.
            self.notify(
                str(exc),
                title="Clipboard error",
                severity="error",
            )

    def compose(self) -> ComposeResult:
        """
        Place components of the app on screen
        :return:
        """
        yield Header()
        log = RichLog(
            id='events',
            highlight=True,
            auto_scroll=True
        )
        log.loading = True
        yield log
        yield Footer()

    async def pump_events(self, log: RichLog):
        """
        Get events from eve log and send them to the application log
        :param log:
        :return:
        """
        try:
            for single_alert in get_events_from_eve(eve_files=self.eve, data_filter=self.data_filter):
                log.loading = False
                log.write(single_alert)
                self.loaded += 1
        except ValueError as ve:
            if hasattr(ve, 'message'):
                reason = ve.message
            elif hasattr(ve, 'reason'):
                reason = f"{ve}"
            else:
                reason = f"{ve}"
            tb = traceback.extract_stack()
            error_screen = ErrorScreen(
                trace=tb,
                reason=reason
            )
            await self.push_screen(error_screen)

    @work(exclusive=False)
    async def on_mount(self):
        """
        Place elements on the screen
        :return:
        """
        log = self.query_one('#events', RichLog)
        await self.pump_events(log)
        if self.loaded > 0:
            self.notify(
                title="Finished loading events",
                severity="information",
                timeout=5,
                message=f"Number of messages loaded: {self.loaded}"
            )



"""
Host data use application
"""
import textwrap
from pathlib import Path
from typing import Type, List

import pyperclip
from textual import work
from textual.app import App, ComposeResult, CSSPathType
from textual.driver import Driver
from textual.widgets import Header, Digits, Footer

from suricatalog import BASEDIR
from suricatalog.clipboard import copy_from_digits
from suricatalog.filter import BaseFilter
from suricatalog.log import get_events_from_eve
from suricatalog.report import HostDataUseReport


class HostDataUse(App):
    """
    Host data usage application
    """
    BINDINGS = [
        ("q", "quit_app", "Quit"),
        ("y,c", "copy_digits", "Copy to clipboard")
    ]
    CSS_PATH = BASEDIR.joinpath('css').joinpath('canned.tcss')
    ENABLE_COMMAND_PALETTE = False

    def __init__(
            self,
            driver_class: Type[Driver] | None = None,
            css_path: CSSPathType | None = None,
            watch_css: bool = False,
            ip_address: str = None,
            data_filter: BaseFilter = None,
            eve: List[Path] = None

    ):
        """
        Constructor
        :param driver_class:
        :param css_path:
        :param watch_css:
        :param ip_address:
        :param data_filter:
        :param eve:
        """
        super().__init__(driver_class, css_path, watch_css)
        self.ip_address = ip_address
        self.data_filter: BaseFilter = data_filter
        self.eve = eve

    def action_quit_app(self) -> None:
        """
        Quit application
        :return:
        """
        self.exit("Exiting Net-Flow now...")

    def action_copy_digits(self) -> None:
        """
        Copy contents to the clipboard
        :return:
        """
        digits = self.query_one(Digits)
        try:
            _, ln = copy_from_digits(digits)
            self.notify(f"Copied {ln} characters!", title="Copied selection")
        except pyperclip.PyperclipException as exc:
            # Show a toast popup if we fail to copy.
            self.notify(
                str(exc),
                title="Clipboard error",
                severity="error",
            )

    def compose(self) -> ComposeResult:
        """
        Place components of the app on screen
        :return:
        """
        yield Header()
        digits = Digits(id="netflow")
        digits.loading = True
        digits.tooltip = textwrap.dedent("""
            Net FLow in bytes.
            """)
        yield digits
        yield Footer()

    @work(exclusive=False)
    async def on_mount(self) -> None:
        """
        Initialize TUI components with data
        :return:
        """
        host_data_user_report = HostDataUseReport()
        for event in get_events_from_eve(
                eve_files=self.eve,
                data_filter=self.data_filter):
            await host_data_user_report.ingest_data(event, self.ip_address)
        digits = self.query_one('#netflow', Digits)
        digits.update(f"{host_data_user_report.bytes:n} bytes")
        digits.loading = False



"""
Screen logic
"""
import sys
from traceback import StackSummary

from textual import on
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Pretty, Button

from suricatalog import BASEDIR


class DetailScreen(ModalScreen):
    """
    Detail screen for an event
    """
    ENABLE_COMMAND_PALETTE = False
    CSS_PATH = BASEDIR.joinpath('css').joinpath('details_screen.tcss')

    def __init__(
            self,
            name: str | None = None,
            ident: str | None = None,
            classes: str | None = None,
            data: str = None
    ):
        """
        Detail screen constructor
        :param name:
        :param ident:
        :param classes:
        :param data:
        """
        super().__init__(name, ident, classes)
        self.data = data

    def compose(self) -> ComposeResult:
        """
        Populate screen component with data
        :return:
        """
        yield Pretty(self.data)
        button = Button("Close", variant="primary", id="close")
        button.tooltip = "Go back to main screen"
        yield button

    @on(Button.Pressed, "#close")
    def on_button_pressed(self, _) -> None:
        """
        Handle exit button
        :param _:
        :return:
        """
        self.app.pop_screen()


class ErrorScreen(ModalScreen):
    """
    Specialized error screen
    """
    ENABLE_COMMAND_PALETTE = False
    CSS_PATH = BASEDIR.joinpath('css').joinpath('details_screen.tcss')

    def __init__(
            self,
            name: str | None = None,
            ident: str | None = None,
            classes: str | None = None,
            data: str = None,
            trace: StackSummary = None,
            reason: str = None
    ):
        """
        Screen constructor
        :param name:
        :param ident:
        :param classes:
        :param data:
        :param trace:
        :param reason:
        """
        super().__init__(name, ident, classes)
        self.data = data
        self.trace = trace
        self.reason = reason

    def compose(self) -> ComposeResult:
        """
        Populate UI components with data
        :return:
        """
        if self.trace and self.reason:
            yield Pretty(
                {
                    'reason': self.reason,
                    'traceback': self.trace
                }
            )
        self.notify(
            title="Unrecoverable error, press 'E' to exit!",
            message="There was an error, application won't recover.",
            severity="error",
            timeout=15
        )
        button = Button.error(
            "Exit application",
            id="close"
        )
        button.tooltip = "Unable to load data, program will exit!"
        yield button

    @on(Button.Pressed, "#close")
    def on_button_pressed(self, _) -> None:
        """
        Handle pressed close button
        :param _:
        :return:
        """
        sys.exit(100)
